import numpy as np
from torch.autograd import Variable
import torch.nn.functional as F


from utils import get_subwindow_tracking


def compute_locations(h, w, stride):
    shifts_x = np.arange(
        0, w * stride, step=stride,
        dtype=np.float32
    )
    shifts_y = np.arange(
        0, h * stride, step=stride,
        dtype=np.float32
    )
    shift_y, shift_x = np.meshgrid(shifts_y, shifts_x)
    shift_x = shift_x.reshape(-1)
    shift_y = shift_y.reshape(-1)
    locations = np.stack((shift_y, shift_x), axis=1) + stride // 2
    return locations


class TrackerConfig(object):
    # These are the default hyper-params for DaSiamRPN 0.3827
    windowing = 'cosine'  # to penalize large displacements [cosine/uniform]
    # Params from the network architecture, have to be consistent with the training
    exemplar_size = 127  # input z size
    instance_size = 255  # input x size (search region)
    total_stride = 8
    score_size = 25
    context_amount = 0.5  # context amount for the exemplar
    location = []
    penalty_k = 0.055
    window_influence = 0.42
    lr = 0.295
    # adaptive change search region #
    adaptive = False
    gpu = True


def tracker_eval(net, x_crop, target_pos, target_sz, window, scale_z, p):
    cls_score, delta, centerness_score = net(x_crop)


    delta = delta.permute(1, 2, 3, 0).contiguous().view(4, -1).data.cpu().numpy()
    cls_score = F.softmax(cls_score.permute(1, 2, 3, 0).contiguous().view(2, -1), dim=0).data[1, :].cpu().numpy()
    centerness_score = F.sigmoid(centerness_score.permute(1, 2, 3, 0).contiguous().view(1, -1)).data.cpu().numpy()

    centerness_score = centerness_score.reshape(-1)
    score = cls_score * centerness_score

    location = p.location.reshape(-1, 2)


    def change(r):
        return np.maximum(r, 1./r)

    def sz(w, h):
        pad = (w + h) * 0.5
        sz2 = (w + pad) * (h + pad)
        return np.sqrt(sz2)

    def sz_wh(wh):
        pad = (wh[0] + wh[1]) * 0.5
        sz2 = (wh[0] + pad) * (wh[1] + pad)
        return np.sqrt(sz2)

    # size penalty
    s_c = change(sz(delta[0, :]+delta[1, :], delta[2, :]+delta[3, :]) / (sz_wh(target_sz)))  # scale penalty
    r_c = change((target_sz[0] / target_sz[1]) / ((delta[0, :]+delta[1, :]) / (delta[2, :]+delta[3, :])))  # ratio penalty

    penalty = np.exp(-(r_c * s_c - 1.) * p.penalty_k)
    pscore = penalty * score * window * window

    best_pscore_id = np.argmax(pscore)

    target = delta[:, best_pscore_id] / scale_z
    target_sz = target_sz / scale_z
    lr = penalty[best_pscore_id] * score[best_pscore_id] * p.lr
    location_target = location[best_pscore_id, :] / scale_z

    res_w = target_sz[0] * (1 - lr) + (target[0] + target[2]) * lr
    res_h = target_sz[1] * (1 - lr) + (target[1] + target[3]) * lr

    # l, t, r, b
    res_x = location_target[0] - target[0] + target_pos[0] + res_w/2
    res_y = location_target[1] - target[1] + target_pos[1] + res_h/2

    target_pos = np.array([res_x, res_y])
    target_sz = np.array([res_w, res_h])
    return target_pos, target_sz, score[best_pscore_id]


def SiamFCOS_init(im, target_pos, target_sz, net, use_gpu=True):
    state = dict()
    p = TrackerConfig()
    if not use_gpu:
        p.gpu = False
    state['im_h'] = im.shape[0]
    state['im_w'] = im.shape[1]

    if p.adaptive:
        if ((target_sz[0] * target_sz[1]) / float(state['im_h'] * state['im_w'])) < 0.004:
            p.instance_size = 255  # small object big search region
        else:
            p.instance_size = 255

        p.score_size = (p.instance_size - p.exemplar_size) / p.total_stride + 1


    h = 25
    w = 25
    stride = p.instance_size / h
    p.location = compute_locations(h, w, stride)

    avg_chans = np.mean(im, axis=(0, 1))

    wc_z = target_sz[0] + p.context_amount * sum(target_sz)
    hc_z = target_sz[1] + p.context_amount * sum(target_sz)
    s_z = round(np.sqrt(wc_z * hc_z))
    z_crop, _ = get_subwindow_tracking(im, target_pos, p.exemplar_size, s_z, avg_chans)

    z = Variable(z_crop.unsqueeze(0))
    if p.gpu:
        z = z.cuda()
    net.temple(z)

    if p.windowing == 'cosine':
        window = np.outer(np.hanning(p.score_size), np.hanning(p.score_size))
    elif p.windowing == 'uniform':
        window = np.ones((int(p.score_size), int(p.score_size)))
    window = window.flatten()

    state['p'] = p
    state['net'] = net
    state['avg_chans'] = avg_chans
    state['window'] = window
    state['target_pos'] = target_pos
    state['target_sz'] = target_sz
    return state


def SiamFCOS_track(state, im):
    p = state['p']
    net = state['net']
    avg_chans = state['avg_chans']
    window = state['window']
    target_pos = state['target_pos']
    target_sz = state['target_sz']

    wc_z = target_sz[1] + p.context_amount * sum(target_sz)
    hc_z = target_sz[0] + p.context_amount * sum(target_sz)
    s_z = np.sqrt(wc_z * hc_z)
    scale_z = p.exemplar_size / s_z
    d_search = (p.instance_size - p.exemplar_size) / 2
    pad = d_search / scale_z
    s_x = s_z + 2 * pad

    a, left_top = get_subwindow_tracking(im, target_pos, p.instance_size, round(s_x), avg_chans)
    x_crop = Variable(a.unsqueeze(0))
    if p.gpu:
        x_crop = x_crop.cuda()
    target_pos, target_sz, score = tracker_eval(net, x_crop, left_top, target_sz * scale_z, window, scale_z, p)
    target_pos[0] = max(0, min(state['im_w'], target_pos[0]))
    target_pos[1] = max(0, min(state['im_h'], target_pos[1]))
    target_sz[0] = max(10, min(state['im_w'], target_sz[0]))
    target_sz[1] = max(10, min(state['im_h'], target_sz[1]))
    state['target_pos'] = target_pos
    state['target_sz'] = target_sz
    state['score'] = score
    return state




